import os
from unittest import defaultTestLoader

import mocker
from paste.fixture import TestApp

from lazr.js.combo import parse_url, combine_files, combo_app


class ComboTestBase(object):

    def makeSampleFile(self, root, fname, content):
        full = os.path.join(root, fname)
        parent = os.path.dirname(full)
        if not os.path.exists(parent):
            os.makedirs(parent)
        return self.makeFile(content=content, path=full)


class ComboTest(ComboTestBase, mocker.MockerTestCase):

    def test_parse_url_keeps_order(self):
        """Parsing a combo loader URL returns an ordered list of filenames."""
        self.assertEquals(
            parse_url(("http://yui.yahooapis.com/combo?"
                       "3.0.0/build/yui/yui-min.js&"
                       "3.0.0/build/oop/oop-min.js&"
                       "3.0.0/build/event-custom/event-custom-min.js&")),
            ("3.0.0/build/yui/yui-min.js",
             "3.0.0/build/oop/oop-min.js",
             "3.0.0/build/event-custom/event-custom-min.js"))

    def test_combine_files_includes_filename(self):
        """Combining files should include their relative filename at the top."""
        test_dir = self.makeDir()

        files = [
            self.makeSampleFile(
                test_dir,
                os.path.join("yui", "yui-min.js"),
                "** yui-min **"),
            self.makeSampleFile(
                test_dir,
                os.path.join("oop", "oop-min.js"),
                "** oop-min **"),
            self.makeSampleFile(
                test_dir,
                os.path.join("event-custom", "event-custom-min.js"),
                "** event-custom-min **"),
            ]

        expected = "\n".join(("// yui/yui-min.js",
                              "** yui-min **",
                              "// oop/oop-min.js",
                              "** oop-min **",
                              "// event-custom/event-custom-min.js",
                              "** event-custom-min **"))
        self.assertEquals(
            "".join(combine_files(["yui/yui-min.js",
                                   "oop/oop-min.js",
                                   "event-custom/event-custom-min.js"],
                                  root=test_dir)).strip(),
            expected)

    def test_combine_css_minifies_and_makes_relative(self):
        """
        Combining CSS files minifies and makes URLs in CSS
        declarations relative to the target path.
        """
        test_dir = self.makeDir()

        files = [
            self.makeSampleFile(
                test_dir,
                os.path.join("widget", "assets", "skins", "sam", "widget.css"),
                """\
                /* widget skin */
                .yui-widget {
                   background: url("img/bg.png");
                }
                """),
            self.makeSampleFile(
                test_dir,
                os.path.join("editor", "assets", "skins", "sam", "editor.css"),
                """\
                /* editor skin */
                .yui-editor {
                   background: url("img/bg.png");
                }
                """),
            ]

        expected = "\n".join(
            ("/* widget/assets/skins/sam/widget.css */",
             ".yui-widget{background:url(widget/assets/skins/sam/img/bg.png)}",
             "/* editor/assets/skins/sam/editor.css */",
             ".yui-editor{background:url(editor/assets/skins/sam/img/bg.png)}",
             ))
        self.assertEquals(
            "".join(combine_files(["widget/assets/skins/sam/widget.css",
                                   "editor/assets/skins/sam/editor.css"],
                                  root=test_dir)).strip(),
            expected)

    def test_missing_file_is_ignored(self):
        """If a missing file is requested we should still combine the others."""
        test_dir = self.makeDir()

        files = [
            self.makeSampleFile(
                test_dir,
                os.path.join("yui", "yui-min.js"),
                "** yui-min **"),
            self.makeSampleFile(
                test_dir,
                os.path.join("event-custom", "event-custom-min.js"),
                "** event-custom-min **"),
            ]

        expected = "\n".join(("// yui/yui-min.js",
                              "** yui-min **",
                              "// oop/oop-min.js",
                              "// [missing]",
                              "// event-custom/event-custom-min.js",
                              "** event-custom-min **"))
        self.assertEquals(
            "".join(combine_files(["yui/yui-min.js",
                                   "oop/oop-min.js",
                                   "event-custom/event-custom-min.js"],
                                  root=test_dir)).strip(),
            expected)

    def test_no_parent_hack(self):
        """If someone tries to hack going up the root, he'll get a miss."""
        test_dir = self.makeDir()
        files = [
            self.makeSampleFile(
                test_dir,
                os.path.join("oop", "oop-min.js"),
                "** oop-min **"),
            ]

        root = os.path.join(test_dir, "root", "lazr")
        os.makedirs(root)

        hack = "../../oop/oop-min.js"
        self.assertTrue(os.path.exists(os.path.join(root, hack)))

        expected = "\n".join(("// ../../oop/oop-min.js",
                              "// [missing]"))
        self.assertEquals(
            "".join(combine_files([hack], root=root)).strip(),
            expected)

    def test_no_absolute_path_hack(self):
        """If someone tries to fetch an absolute file, he'll get nothing."""
        test_dir = self.makeDir()

        hack = "/etc/passwd"
        self.assertTrue(os.path.exists("/etc/passwd"))

        expected = ""
        self.assertEquals(
            "".join(combine_files([hack], root=test_dir)).strip(),
            expected)


class WSGIComboTest(ComboTestBase, mocker.MockerTestCase):

    def setUp(self):
        self.root = self.makeDir()
        self.app = TestApp(combo_app(self.root))

    def test_combo_app_sets_content_type_for_js(self):
        """The WSGI App should set a proper Content-Type for Javascript."""
        files = [
            self.makeSampleFile(
                self.root,
                os.path.join("yui", "yui-min.js"),
                "** yui-min **"),
            self.makeSampleFile(
                self.root,
                os.path.join("oop", "oop-min.js"),
                "** oop-min **"),
            self.makeSampleFile(
                self.root,
                os.path.join("event-custom", "event-custom-min.js"),
                "** event-custom-min **"),
            ]

        expected = "\n".join(("// yui/yui-min.js",
                              "** yui-min **",
                              "// oop/oop-min.js",
                              "** oop-min **",
                              "// event-custom/event-custom-min.js",
                              "** event-custom-min **"))

        res = self.app.get("/?" + "&".join(
            ["yui/yui-min.js",
             "oop/oop-min.js",
             "event-custom/event-custom-min.js"]), status=200)
        self.assertEquals(res.headers, [("Content-Type", "text/javascript")])
        self.assertEquals(res.body.strip(), expected)

    def test_combo_app_sets_content_type_for_css(self):
        """The WSGI App should set a proper Content-Type for CSS."""
        files = [
            self.makeSampleFile(
                self.root,
                os.path.join("widget", "skin", "sam", "widget.css"),
                "/* widget-skin-sam */"),
            ]

        expected = "/* widget/skin/sam/widget.css */"

        res = self.app.get("/?" + "&".join(
            ["widget/skin/sam/widget.css"]), status=200)
        self.assertEquals(res.headers, [("Content-Type", "text/css")])
        self.assertEquals(res.body.strip(), expected)

    def test_no_filename_gives_404(self):
        """If no filename is included, a 404 should be returned."""
        res = self.app.get("/", status=404)
        self.assertEquals(res.headers, [("Content-Type", "text/plain")])
        self.assertEquals(res.body, "Not Found")


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
